#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-09-11
# @Filename: version.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from coordio import __version__ as coordio_version

from cherno import __version__ as cherno_version

from . import ChernoCommandType, cherno_parser


__all__ = ["version"]


@cherno_parser.command()
async def version(command: ChernoCommandType):
    """Prints the current version of cherno and dependencies."""

    command.info(version=cherno_version)
    command.info(coordio_version=coordio_version)

    try:
        from astrometry import __version__ as astrometrynet_version  # type: ignore

        command.info(astrometrynet_version=astrometrynet_version)
    except ImportError:
        command.warning("Cannot find astrometry.net version.")

    try:
        from fps_calibrations import get_version

        command.info(fps_calibrations_version=get_version())
    except ImportError:
        command.warning("Cannot find fps_calibrations version.")

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-02-07
# @Filename: acquisition.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib
import time
import warnings
from dataclasses import dataclass, field
from functools import partial

from typing import TYPE_CHECKING, Any, Coroutine

import numpy
from astropy.io import fits
from astropy.table import Table
from astropy.wcs import WCS, FITSFixedWarning
from simple_pid.PID import PID

from clu.command import FakeCommand
from coordio.astrometry import AstrometryNet
from coordio.guide import GuiderFitter, gfa_to_radec

from cherno import config, log
from cherno.exceptions import ChernoError
from cherno.extraction import Extraction, ExtractionData, PathLike
from cherno.lcotcc import apply_correction_lco
from cherno.maskbits import GuiderStatus
from cherno.tcc import apply_axes_correction, apply_focus_correction
from cherno.utils import focus_fit, run_in_executor


if TYPE_CHECKING:
    from cherno.actor import ChernoActor, ChernoCommandType
    from cherno.actor.actor import ChernoState


warnings.filterwarnings("ignore", module="astropy.wcs.wcs")
warnings.filterwarnings("ignore", category=FITSFixedWarning)


__all__ = ["Acquisition"]


@dataclass
class AcquisitionData:
    """Data from the acquisition process."""

    camera: str
    extraction_data: ExtractionData
    solved: bool = False
    solve_time: float = -999.0
    wcs: WCS = field(default_factory=WCS)
    camera_racen: float = -999.0
    camera_deccen: float = -999.0
    xrot: float = -999.0
    yrot: float = -999.0
    rotation: float = -999.0
    proc_image: pathlib.Path | None = None

    def __post_init__(self):

        self.path = self.extraction_data.path
        self.camera = self.extraction_data.camera
        self.exposure_no = self.extraction_data.exposure_no
        self.obstime = self.extraction_data.obstime
        self.observatory = self.extraction_data.observatory
        self.field_ra = self.extraction_data.field_ra
        self.field_dec = self.extraction_data.field_dec
        self.field_pa = self.extraction_data.field_pa

    @property
    def e_data(self):
        """Shortcut for ``extraction_data``."""

        return self.extraction_data


@dataclass
class AstrometricSolution:
    """Astrometric solution data."""

    valid_solution: bool
    acquisition_data: list[AcquisitionData]
    delta_ra: float = -999.0
    delta_dec: float = -999.0
    delta_rot: float = -999.0
    delta_scale: float = -999.0
    delta_focus: float = -999.0
    fit_mode: str = "full"
    rms: float = -999.0
    fwhm_median: float = -999.0
    fwhm_fit: float = -999.0
    focus_coeff: list[float] = field(default_factory=lambda: [-999.0] * 3)
    focus_r2: float = -999.0
    camera_rotation: float = -999.0
    correction_applied: list[float] = field(default_factory=lambda: [0.0] * 5)


class AxesPID:
    """Store for the axis PID coefficient."""

    def __init__(self, actor: ChernoActor | None = None):

        self.actor = actor

        self.ra = self.reset("ra")
        self.dec = self.reset("dec")
        self.rot = self.reset("rot")
        self.focus = self.reset("focus")

    def reset(self, axis: str):
        """Restart the PID loop for an axis."""

        if self.actor is None:
            pid_coeffs = config["guide_loop"][axis]["pid"]
        else:
            pid_coeffs = self.actor.state.guide_loop[axis]["pid"]

        return PID(
            Kp=pid_coeffs["k"],
            Ki=pid_coeffs.get("ti", 0),
            Kd=pid_coeffs.get("td", 0),
        )


class Acquisition:
    """Runs the steps for field acquisition."""

    def __init__(
        self,
        observatory: str,
        extractor: Extraction | None = None,
        astrometry: AstrometryNet | None = None,
        command: ChernoCommandType | None = None,
        astrometry_params: dict = {},
        extraction_params: dict = {},
    ):

        self.extractor = extractor or Extraction(observatory, **extraction_params)

        if astrometry is not None:
            self.astrometry = astrometry
        else:
            pixel_scale = config["pixel_scale"]
            astrometry_net_config = config["acquisition"]["astrometry_net_config"]
            backend_config = pathlib.Path(__file__).parent / astrometry_net_config
            self.astrometry = AstrometryNet(
                backend_config=str(backend_config),
                width=2048,
                height=2048,
                no_plots=True,
                scale_low=pixel_scale * 0.9,
                scale_high=pixel_scale * 1.1,
                scale_units="arcsecperpix",
                sort_column="flux",
                radius=0.5,
                cpulimit=config["acquisition"]["cpulimit"],
                **astrometry_params,
            )

        self.observatory = observatory.upper()
        self.fitter = GuiderFitter(self.observatory)
        self.command = command or FakeCommand(log)
        self.pids = AxesPID(command.actor if command is not None else None)

    def set_command(self, command: ChernoCommandType):
        """Sets the command."""

        self.command = command

    async def process(
        self,
        command: ChernoCommandType | None,
        images: PathLike | list[PathLike],
        write_proc: bool = True,
        overwrite: bool = False,
        correct: bool = True,
        full_correction: bool = False,
        offset: list[float] | None = None,
        scale_rms: bool = True,
        wait_for_correction: bool = True,
        only_radec: bool = False,
        auto_radec_min: int = 2,
    ):
        """Performs extraction and astrometry."""

        if command is not None:
            self.set_command(command)

        if not isinstance(images, (list, tuple)):
            images = [images]

        self.command.info("Extracting sources.")

        ext_data = await asyncio.gather(
            *[
                run_in_executor(self.extractor.process, im, executor="process")
                for im in images
            ]
        )

        for d in ext_data:
            if d.nvalid == 0:
                self.command.warning(f"Camera {d.camera}: not enough sources.")
            else:
                self.command.info(
                    fwhm_camera=[
                        d.camera,
                        d.exposure_no,
                        d.fwhm_median,
                        d.nregions,
                        d.nvalid,
                    ]
                )

        self.command.info("Running astrometry.net.")
        acq_data = await asyncio.gather(*[self._astrometry_one(d) for d in ext_data])

        self.command.debug("Saving proc- file.")
        if write_proc:
            await asyncio.gather(
                *[self.write_proc_image(d, overwrite=overwrite) for d in acq_data]
            )

        ast_solution = await self.fit(
            list(acq_data),
            offset=offset,
            scale_rms=scale_rms,
            only_radec=only_radec,
            auto_radec_min=auto_radec_min,
        )

        if correct and ast_solution.valid_solution is True:
            await self.correct(
                ast_solution,
                full=full_correction,
                wait_for_correction=wait_for_correction,
            )
        else:
            self.command.info(
                acquisition_valid=ast_solution.valid_solution,
                did_correct=any(ast_solution.correction_applied),
                correction_applied=ast_solution.correction_applied,
            )

        if self.command.actor:
            update_proc_headers(ast_solution, self.command.actor.state)

        return ast_solution

    async def fit(
        self,
        data: list[AcquisitionData],
        offset: list[float] | None = None,
        scale_rms: bool = False,
        do_focus: bool = True,
        only_radec: bool = False,
        auto_radec_min: int = 2,
    ):
        """Calculate the astrometric solution."""

        ast_solution = AstrometricSolution(False, data)

        solved = sorted([d for d in data if d.solved is True], key=lambda x: x.camera)
        weights = [s.extraction_data.nvalid for s in solved]

        if len(solved) == 0:
            self.command.error(acquisition_valid=False, did_correct=False)
            return ast_solution

        fwhm = [d.e_data.fwhm_median for d in data if d.e_data.fwhm_median > 0]
        fwhm = numpy.average(fwhm)

        camera_rotation = numpy.average([d.rotation for d in solved], weights=weights)

        ast_solution.fwhm_median = numpy.round(float(fwhm), 3)
        ast_solution.camera_rotation = numpy.round(float(camera_rotation), 2)

        if solved[0].field_ra == "NaN" or isinstance(solved[0].field_ra, str):
            self.command.error(acquisition_valid=False, did_correct=False)
            self.command.error("Field not defined. Cannot run astrometric fit.")
            return ast_solution

        if offset is None:
            if self.command.actor:
                offset = list(self.command.actor.state.offset)
            else:
                offset = list(config.get("offset", [0.0, 0.0, 0.0]))

        self.fitter.reset()
        for d in solved:
            self.fitter.add_wcs(d.camera, d.wcs, d.obstime.jd)

        field_ra = solved[0].field_ra
        field_dec = solved[0].field_dec
        field_pa = solved[0].field_pa

        default_offset = config.get("default_offset", (0.0, 0.0, 0.0))
        full_offset = numpy.array(offset) + numpy.array(default_offset)

        self.command.debug(default_offset=default_offset)
        if any(offset):
            self.command.warning(offset=offset)
        else:
            self.command.debug(offset=offset)

        if only_radec is True:
            self.command.warning(
                "Only fitting RA/Dec. The rotation and scale offsets "
                "are informational-only and not corrected."
            )
        elif auto_radec_min >= 0 and len(solved) <= auto_radec_min:
            only_radec = True
            self.command.warning(
                f"Only {len(solved)} cameras solved. Only fitting RA/Dec. "
                "The rotation and scale offsets are informational-only "
                "and not corrected."
            )

        guide_fit = self.fitter.fit(
            field_ra,
            field_dec,
            field_pa,
            offset=full_offset,
            scale_rms=scale_rms,
            only_radec=only_radec,
        )

        exp_no = solved[0].exposure_no  # Should be the same for all.

        if guide_fit is False:
            rms = delta_ra = delta_dec = delta_rot = delta_scale = -999.0
        else:
            delta_ra = guide_fit.delta_ra
            delta_dec = guide_fit.delta_dec
            delta_rot = guide_fit.delta_rot
            delta_scale = guide_fit.delta_scale

            xrms = guide_fit.xrms
            yrms = guide_fit.yrms
            rms = guide_fit.rms

            self.command.info(guide_rms=[exp_no, xrms, yrms, rms])

        ast_solution.valid_solution = True
        ast_solution.delta_ra = float(delta_ra)
        ast_solution.delta_dec = float(delta_dec)
        ast_solution.delta_rot = float(delta_rot)
        ast_solution.delta_scale = float(delta_scale)
        ast_solution.rms = float(rms)

        if guide_fit.only_radec:
            ast_solution.fit_mode = "radec"

        if do_focus:
            try:
                fwhm_fit, x_min, a, b, c, r2 = focus_fit(
                    [d.e_data for d in data],
                    plot=config["acquisition"]["plot_focus"],
                )

                # Relationship between M2 move and focal plane. See
                # http://www.loptics.com/ATM/mirror_making/cass_info/cass_info.html
                focus_sensitivity = config["focus_sensitivity"]

                ast_solution.fwhm_fit = round(fwhm_fit, 3)
                ast_solution.delta_focus = round(-x_min / focus_sensitivity, 1)
                ast_solution.focus_coeff = [a, b, c]
                ast_solution.focus_r2 = round(r2, 3)

            except Exception as err:
                self.command.warning(f"Failed fitting focus curve: {err}.")

        self.command.info(
            astrometry_fit=[
                exp_no,
                len(solved),
                -999.0,
                -999.0,
                numpy.round(fwhm, 2),
                -999,
                numpy.round(camera_rotation, 3),
                delta_ra,
                delta_dec,
                delta_rot,
                delta_scale,
            ]
        )

        # Output focus data in a single keyword, mostly for Boson's benefit.
        focus_data = [int(exp_no)]
        for d in ast_solution.acquisition_data:
            if d.extraction_data.fwhm_median > 0 and d.extraction_data.nvalid > 0:
                focus_data += [
                    int(d.camera[-1]),
                    d.extraction_data.fwhm_median,
                    d.extraction_data.focus_offset,
                ]

        self.command.debug(focus_data=focus_data)

        self.command.info(
            focus_fit=[
                exp_no,
                ast_solution.fwhm_fit,
                float(f"{ast_solution.focus_coeff[0]:.3e}"),
                float(f"{ast_solution.focus_coeff[1]:.3e}"),
                float(f"{ast_solution.focus_coeff[2]:.3e}"),
                ast_solution.focus_r2,
                ast_solution.delta_focus,
            ]
        )

        if (
            delta_scale > 0
            and self.command.actor
            and fwhm < 2.5
            and not guide_fit.only_radec
        ):
            # If we measured the scale, add it to the actor state. This is later
            # used to compute the average scale over a period. We also add the time
            # because we'll want to reject measurements that are too old.
            self.command.actor.state.scale_history.append((time.time(), delta_scale))

        return ast_solution

    async def correct(
        self,
        data: AstrometricSolution,
        full: bool = False,
        wait_for_correction: bool = True,
    ):
        """Runs the astrometric fit"""

        if not self.command.actor or self.command.status.is_done:
            raise ChernoError("Cannot run correct without a valid running command.")

        actor_state = self.command.actor.state
        guider_status = actor_state.status

        stopping = guider_status & (GuiderStatus.STOPPING | GuiderStatus.IDLE)

        if stopping:
            data.correction_applied = [0.0, 0.0, 0.0, 0.0, 0.0]
            self.command.info(acquisition_valid=True, did_correct=False)
            return True

        self.command.info("Applying corrections.")

        if self.observatory == "APO":
            await self._correct_apo(data, full=full)
        else:
            await self._correct_lco(
                data,
                full=full,
                wait_for_correction=wait_for_correction,
            )

    async def _correct_apo(self, data: AstrometricSolution, full: bool = False):

        actor_state = self.command.actor.state

        min_isolated = actor_state.guide_loop["rot"]["min_isolated_correction"]
        if abs(data.delta_rot) >= min_isolated:
            self.command.debug("Applying only large rotator correction.")
            coro = apply_axes_correction(
                self.command,
                self.pids,
                delta_rot=data.delta_rot if data.fit_mode == "full" else None,
                full=full,
            )

        else:
            coro = apply_axes_correction(
                self.command,
                self.pids,
                delta_radec=(data.delta_ra, data.delta_dec),
                delta_rot=data.delta_rot if data.fit_mode == "full" else None,
                full=full,
            )

        correct_tasks: list[Coroutine[Any, Any, Any]] = [coro]

        do_focus: bool = False

        # Ignore focus correction when the r2 correlation is bad or when we got
        # an inverted parabola.
        if (
            data.focus_r2 > config["acquisition"]["focus_r2_threshold"]
            and data.focus_coeff[0] > 0
        ):
            do_focus = True
            coro = apply_focus_correction(self.command, self.pids, data.delta_focus)
            correct_tasks.append(coro)
        else:
            self.command.warning("Focus fit poorly constrained. Not correcting focus.")

        self.command.actor.state.set_status(GuiderStatus.CORRECTING, mode="add")
        applied_corrections: Any = await asyncio.gather(*correct_tasks)
        self.command.actor.state.set_status(GuiderStatus.CORRECTING, mode="remove")

        data.correction_applied[:4] = applied_corrections[0]
        if do_focus:
            data.correction_applied[4] = applied_corrections[1] or 0.0
        else:
            data.correction_applied[4] = 0.0

        self.command.info(
            acquisition_valid=True,
            did_correct=any(data.correction_applied),
            correction_applied=data.correction_applied,
        )

    async def _correct_lco(
        self,
        data: AstrometricSolution,
        full: bool = False,
        wait_for_correction: bool = True,
    ):

        do_focus: bool = False

        enabled_axes = self.command.actor.state.enabled_axes

        # Ignore focus correction when the r2 correlation is bad or when we got
        # an inverted parabola.
        if (
            data.focus_r2 > config["acquisition"]["focus_r2_threshold"]
            and data.focus_coeff[0] > 0
            and "focus" in enabled_axes
        ):
            do_focus = True
        else:
            self.command.warning("Focus fit poorly constrained. Not correcting focus.")

        self.command.actor.state.set_status(GuiderStatus.CORRECTING, mode="add")

        applied_corrections: Any = await apply_correction_lco(
            self.command,
            self.pids,
            delta_radec=(data.delta_ra, data.delta_dec),
            delta_rot=data.delta_rot if data.fit_mode == "full" else None,
            delta_focus=data.delta_focus if do_focus else None,
            full=full,
            wait_for_correction=wait_for_correction,
        )

        self.command.actor.state.set_status(GuiderStatus.CORRECTING, mode="remove")

        data.correction_applied = applied_corrections

        self.command.info(
            acquisition_valid=True,
            did_correct=any(data.correction_applied),
            correction_applied=data.correction_applied,
        )

    async def _astrometry_one(self, ext_data: ExtractionData):

        if config["acquisition"]["astrometry_net_use_all_regions"]:
            regions = ext_data.regions.copy()
        else:
            regions = ext_data.regions.loc[ext_data.regions.valid == 1].copy()

        if self.astrometry._options.get("dir", None) is None:
            astrometry_dir = pathlib.Path(config["acquisition"]["astrometry_dir"])
            if astrometry_dir.is_absolute():
                pass
            else:
                astrometry_dir = ext_data.path.parent / astrometry_dir

            astrometry_dir.mkdir(exist_ok=True, parents=True)
        else:
            astrometry_dir = pathlib.Path(self.astrometry._options["dir"])

        outfile_root = astrometry_dir / ext_data.path.stem

        # We use all detections, even invalid ones here.
        if ext_data.algorithm == "marginal":
            xyls_df = regions.loc[:, ["x", "y", "flux"]].copy()
            xyls_df = xyls_df.rename(columns={"x_fit": "x", "y_fit": "y"})
        elif ext_data.algorithm == "daophot":
            xyls_df = regions.loc[:, ["x_0", "y_0", "flux_0"]].copy()
            xyls_df = xyls_df.rename(columns={"x_0": "x", "y_0": "y", "flux_0": "flux"})
        else:
            xyls_df = regions.loc[:, ["x", "y", "flux"]]

        gfa_xyls = Table.from_pandas(xyls_df)
        gfa_xyls_path = outfile_root.with_suffix(".xyls")
        gfa_xyls.write(str(gfa_xyls_path), format="fits", overwrite=True)

        wcs_output = outfile_root.with_suffix(".wcs")
        wcs_output.unlink(missing_ok=True)

        camera_id = int(ext_data.camera[-1])
        radec_centre = gfa_to_radec(
            ext_data.observatory,
            1024,
            1024,
            camera_id,
            ext_data.field_ra,
            ext_data.field_dec,
            position_angle=ext_data.field_pa,
        )

        proc = await self.astrometry.run_async(
            [gfa_xyls_path],
            stdout=outfile_root.with_suffix(".stdout"),
            stderr=outfile_root.with_suffix(".stderr"),
            ra=radec_centre[0],
            dec=radec_centre[1],
            odds_to_solve=10**self.command.actor.state.astrometry_net_odds,
        )

        acq_data = AcquisitionData(ext_data.camera, ext_data, solve_time=proc.elapsed)

        if wcs_output.exists():
            acq_data.solved = True
            wcs = WCS(open(wcs_output).read())
            acq_data.wcs = wcs

            racen, deccen = wcs.pixel_to_world_values([[1024, 1024]])[0]
            acq_data.camera_racen = float(numpy.round(racen, 6))
            acq_data.camera_deccen = float(numpy.round(deccen, 6))

            # TODO: consider parallactic angle here.
            cd: numpy.ndarray = wcs.wcs.cd
            rot_rad = numpy.arctan2(
                numpy.array([cd[0, 1], cd[0, 0]]),
                numpy.array([cd[1, 1], cd[1, 0]]),
            )

            # Rotation is from N to E to the x and y axes of the GFA.
            yrot, xrot = numpy.rad2deg(rot_rad)
            acq_data.xrot = float(numpy.round(xrot % 360.0, 3))
            acq_data.yrot = float(numpy.round(yrot % 360.0, 3))

            # Calculate field rotation.
            cameras = config["cameras"]
            camera_rot = cameras["rotation"][acq_data.camera]
            rotation = numpy.array([xrot - camera_rot - 90, yrot - camera_rot]) % 360
            rotation[rotation > 180.0] -= 360.0
            rotation = numpy.mean(rotation)
            acq_data.rotation = float(numpy.round(rotation, 3))

        self.command.info(
            camera_solution=[
                acq_data.camera,
                acq_data.exposure_no,
                acq_data.solved,
                acq_data.camera_racen,
                acq_data.camera_deccen,
                acq_data.xrot,
                acq_data.yrot,
                acq_data.rotation,
            ]
        )

        return acq_data

    async def write_proc_image(
        self,
        acq_data: AcquisitionData,
        outpath: PathLike = None,
        overwrite: bool = False,
    ):
        """Writes the proc-gimg image."""

        ext_data = acq_data.extraction_data

        proc_hdu = fits.open(str(acq_data.path)).copy()

        rec = Table.from_pandas(acq_data.extraction_data.regions).as_array()
        proc_hdu.append(fits.BinTableHDU(rec, name="CENTROIDS"))

        proc_hdu[1].header["SOLVED"] = acq_data.solved
        proc_hdu[1].header["SOLVTIME"] = (
            acq_data.solve_time,
            "Time to solve the field or fail",
        )

        if self.command.actor is not None:
            offsets = self.command.actor.state.offset
        else:
            offsets = [-999.0] * 3

        proc_hdu[1].header["OFFRA"] = (offsets[0], "Relative offset in RA [arcsec]")
        proc_hdu[1].header["OFFDEC"] = (offsets[1], "Relative offset in Dec [arcsec]")
        proc_hdu[1].header["OFFPA"] = (offsets[2], "Relative offset in PA [arcsec]")

        default_offset = config.get("default_offset", (0.0, 0.0, 0.0))
        aoffset = (
            default_offset[0] + offsets[0],
            default_offset[1] + offsets[1],
            default_offset[2] + offsets[2],
        )
        proc_hdu[1].header["AOFFRA"] = (aoffset[0], "Absolute offset in RA [arcsec]")
        proc_hdu[1].header["AOFFDEC"] = (aoffset[1], "Absolute offset in Dec [arcsec]")
        proc_hdu[1].header["AOFFPA"] = (aoffset[2], "Absolute offset in PA [arcsec]")

        proc_hdu[1].header.update(acq_data.wcs.to_header())

        proc_path: pathlib.Path
        if outpath is not None:
            proc_path = pathlib.Path(outpath)
            if proc_path.is_dir():
                proc_path = proc_path / ("proc-" + ext_data.path.name)
        else:
            proc_path = ext_data.path.parent / ("proc-" + ext_data.path.name)

        loop = asyncio.get_running_loop()
        func = partial(
            proc_hdu.writeto,
            proc_path,
            overwrite=overwrite,
            output_verify="silentfix",
        )
        await loop.run_in_executor(None, func)

        acq_data.proc_image = proc_path


def update_proc_headers(data: AstrometricSolution, guider_state: ChernoState):

    guide_loop = guider_state.guide_loop

    enabled_axes = guider_state.enabled_axes
    enabled_ra = "ra" in enabled_axes
    enabled_dec = "dec" in enabled_axes
    enabled_rot = "rot" in enabled_axes
    enabled_focus = "focus" in enabled_axes

    cra, cdec, crot, cscl, cfoc = data.correction_applied

    ra_pid_k = guide_loop["ra"]["pid"]["k"]
    ra_pid_td = guide_loop["ra"]["pid"].get("td", 0.0)
    ra_pid_ti = guide_loop["ra"]["pid"].get("ti", 0.0)

    dec_pid_k = guide_loop["dec"]["pid"]["k"]
    dec_pid_td = guide_loop["dec"]["pid"].get("td", 0.0)
    dec_pid_ti = guide_loop["dec"]["pid"].get("ti", 0.0)

    rot_pid_k = guide_loop["rot"]["pid"]["k"]
    rot_pid_td = guide_loop["rot"]["pid"].get("td", 0.0)
    rot_pid_ti = guide_loop["rot"]["pid"].get("ti", 0.0)

    focus_pid_k = guide_loop["focus"]["pid"]["k"]
    focus_pid_td = guide_loop["focus"]["pid"].get("td", 0.0)
    focus_pid_ti = guide_loop["focus"]["pid"].get("ti", 0.0)

    if "scale" in guide_loop:
        scale_pid_k = guide_loop["scale"]["pid"]["k"]
        scale_pid_td = guide_loop["scale"]["pid"].get("td", 0.0)
        scale_pid_ti = guide_loop["scale"]["pid"].get("ti", 0.0)
    else:
        scale_pid_k = 0.0
        scale_pid_td = 0.0
        scale_pid_ti = 0.0

    rms = data.rms
    delta_ra = data.delta_ra
    delta_dec = data.delta_dec
    delta_rot = data.delta_rot
    delta_scale = data.delta_scale
    delta_focus = data.delta_focus

    # Update headers of proc images with deltas.
    for a_data in data.acquisition_data:
        if a_data.proc_image is not None:
            hdus = fits.open(str(a_data.proc_image), mode="update")
            header = hdus[1].header

            header["EXTMETH"] = (a_data.e_data.algorithm, "Algorithm for star finding")

            header["RAK"] = (ra_pid_k, "PID K term for RA")
            header["RATD"] = (ra_pid_td, "PID Td term for RA")
            header["RATI"] = (ra_pid_ti, "PID Ti term for RA")

            header["DECK"] = (dec_pid_k, "PID K term for Dec")
            header["DECTD"] = (dec_pid_td, "PID Td term for Dec")
            header["DECTI"] = (dec_pid_ti, "PID Ti term for Dec")

            header["ROTK"] = (rot_pid_k, "PID K term for Rot.")
            header["ROTTD"] = (rot_pid_td, "PID Td term for Rot.")
            header["ROTTI"] = (rot_pid_ti, "PID Ti term for Rot.")

            header["SCLK"] = (scale_pid_k, "PID K term for Scale")
            header["SCLTD"] = (scale_pid_td, "PID Td term for Scale")
            header["SCLTI"] = (scale_pid_ti, "PID Ti term for Scale")

            header["FOCUSK"] = (focus_pid_k, "PID K term for Focus")
            header["FOCUSTD"] = (focus_pid_td, "PID Td term for Focus")
            header["FOCUSTI"] = (focus_pid_ti, "PID Ti term for Focus")

            header["FWHM"] = (a_data.e_data.fwhm_median, "Mesured FWHM [arcsec]")
            header["FWHMFIT"] = (data.fwhm_fit, "Fitted FWHM [arcsec]")

            header["RMS"] = (rms, "Guide RMS [arcsec]")
            header["FITMODE"] = (data.fit_mode, "Fit mode (full or RA/Dec)")

            header["E_RA"] = (enabled_ra, "RA corrections enabled?")
            header["E_DEC"] = (enabled_dec, "Dec corrections enabled?")
            header["E_ROT"] = (enabled_rot, "Rotator corrections enabled?")
            header["E_FOCUS"] = (enabled_focus, "Focus corrections enabled?")
            header["E_SCL"] = (False, "Scale corrections enabled?")

            header["DELTARA"] = (delta_ra, "RA measured delta [arcsec]")
            header["DELTADEC"] = (delta_dec, "Dec measured delta [arcsec]")
            header["DELTAROT"] = (delta_rot, "Rotator measured delta [arcsec]")
            header["DELTASCL"] = (delta_scale, "Scale measured factor")
            header["DELTAFOC"] = (delta_focus, "Focus delta [microns]")

            header["CORR_RA"] = (cra, "RA applied correction [arcsec]")
            header["CORR_DEC"] = (cdec, "Dec applied correction [arcsec]")
            header["CORR_ROT"] = (crot, "Rotator applied correction [arcsec]")
            header["CORR_SCL"] = (cscl, "Scale applied correction")
            header["CORR_FOC"] = (cfoc, "Focus applied correction [microns]")

            hdus.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-12
# @Filename: tcc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy

from cherno.exceptions import ChernoError


if TYPE_CHECKING:
    from cherno.acquisition import AxesPID
    from cherno.actor import ChernoCommandType


__all__ = ["apply_correction_lco"]


async def apply_correction_lco(
    command: ChernoCommandType,
    pids: AxesPID,
    delta_radec: tuple[float, float] | numpy.ndarray | None = None,
    delta_rot: float | None = None,
    delta_focus: float | None = None,
    full: bool = False,
    wait_for_correction: bool = True,
):
    """Send corrections to the LCOTCC. Corrections here are in arcsec."""

    assert command.actor

    state = command.actor.state

    guide_loop = state.guide_loop
    enabled_axes = state.enabled_axes

    corr_radec = numpy.array([0.0, 0.0])
    corr_rot: float = 0.0
    corr_focus: float = 0.0

    if delta_rot is not None and "rot" in enabled_axes:

        if full:
            corr_rot = -delta_rot
        else:
            corr_rot = pids.rot(delta_rot) or 0.0

        min_corr_arcsec = guide_loop["rot"]["min_correction"]
        max_corr_arcsec = guide_loop["rot"]["max_correction"]

        if numpy.abs(corr_rot) < min_corr_arcsec:
            command.debug("Skipping small rotator correction.")
            corr_rot = 0.0
        elif numpy.abs(corr_rot) > max_corr_arcsec:
            raise ChernoError(f"Rotator correction too large: {corr_rot:.1f}.")

        corr_rot /= 3600

    if corr_rot == 0.0 and delta_radec is not None:

        for ax_idx, ax in enumerate(["ra", "dec"]):
            if ax not in enabled_axes:
                continue

            if full:
                corr_ax = -float(delta_radec[ax_idx])
            else:
                # This returns the correction (i.e., opposite sign).
                corr_ax = getattr(pids, ax)(float(delta_radec[ax_idx])) or 0.0

            min_corr_arcsec = guide_loop[ax]["min_correction"]
            max_corr_arcsec = guide_loop[ax]["max_correction"]

            if numpy.abs(corr_ax) < min_corr_arcsec:
                command.debug(f"Skipping small {ax.lower()} correction.")
                corr_ax = 0.0
            elif numpy.abs(corr_ax) > max_corr_arcsec:
                raise ChernoError(f"{ax.upper()} correction too large: {corr_ax:.2f}.")

            corr_radec[ax_idx] = corr_ax / 3600

    if delta_focus is not None and "focus" in enabled_axes:

        if full:
            corr_focus = -delta_focus
        else:
            corr_focus = pids.focus(delta_focus) or 0.0

        min_corr = guide_loop["focus"]["min_correction"]
        max_corr = guide_loop["focus"]["max_correction"]

        if numpy.abs(corr_focus) < min_corr:
            command.debug("Skipping small focus correction.")
            corr_focus = 0.0

        elif numpy.abs(corr_focus) > max_corr:
            command.warning(f"Skipping large focus correction: {corr_focus:.1f}.")
            corr_focus = 0.0

    # Correction applied in ra, dec, rot, scale, and focus.
    correction_applied = [0.0, 0.0, 0.0, 0.0, 0.0]

    if (
        numpy.any(numpy.abs(corr_radec) > 0)
        or numpy.abs(corr_rot) > 0
        or numpy.abs(corr_focus) > 0
    ):
        wait_flag = "" if wait_for_correction is True else "/waittime=0"
        tcc_offset_cmd = await command.send_command(
            "lcotcc",
            f"guideoffset {corr_radec[0]},{corr_radec[1]},{corr_rot},{corr_focus} "
            f"{wait_flag}",
        )

        if tcc_offset_cmd.status.did_fail:
            command.error("Failed applying RA/Dec correction.")
            return correction_applied

        correction_applied[0:2] = numpy.round(corr_radec * 3600.0, 3)
        correction_applied[2] = numpy.round(corr_rot * 3600.0, 3)
        correction_applied[4] = numpy.round(corr_focus, 1)

    return correction_applied



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-13
# @Filename: offset.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import click

from cherno import config

from . import ChernoCommandType, cherno_parser


__all__ = ["offset"]


@cherno_parser.command()
@click.argument("RA", type=float, required=False)
@click.argument("DEC", type=float, required=False)
@click.argument("PA", type=float, required=False, default=0.0)
async def offset(
    command: ChernoCommandType,
    ra: float | None = None,
    dec: float | None = None,
    pa: float = 0.0,
):
    """Offsets the field boresight by RA/DEC/PA arcsec."""

    if ra is None and dec is None:
        command.warning("Resetting offset.")
        ra = 0.0
        dec = 0.0
        pa = 0.0
    elif ra is not None and dec is None:
        return command.fail(error="ra and dec are required.")

    assert ra is not None and dec is not None and pa is not None

    assert command.actor
    command.actor.state.offset = (ra, dec, pa)

    default_offset = config.get("default_offset", (0.0, 0.0, 0.0))

    return command.finish(offset=[ra, dec, pa], default_offset=default_offset)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-02-07
# @Filename: extraction.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import multiprocessing
import pathlib
import re
import warnings
from copy import deepcopy
from dataclasses import dataclass, field

from typing import Any, TypeVar, cast

import matplotlib.pyplot as plt
import numpy
import pandas
import seaborn
import sep
from astropy.io import fits
from astropy.modeling.fitting import LevMarLSQFitter
from astropy.modeling.models import Gaussian1D, Trapezoid1D
from astropy.stats import SigmaClip, gaussian_sigma_to_fwhm
from astropy.table import Table
from astropy.time import Time, TimeDelta
from matplotlib.patches import Ellipse
from photutils.background import MedianBackground, StdBackgroundRMS
from photutils.detection import DAOStarFinder
from photutils.psf import BasicPSFPhotometry, DAOGroup, IntegratedGaussianPRF

from cherno import config


PathLike = TypeVar("PathLike", pathlib.Path, str)
seaborn.set_theme(style="white")


__all__ = ["ExtractionData", "Extraction"]


@dataclass
class ExtractionData:
    """Data from extraction."""

    image: str
    path: pathlib.Path
    camera: str
    exposure_no: int
    mjd: int
    obstime: Time
    observatory: str
    field_ra: float
    field_dec: float
    field_pa: float
    algorithm: str
    regions: pandas.DataFrame = field(default_factory=pandas.DataFrame)
    nregions: int = 0
    nvalid: int = 0
    fwhm_median: float = -999.0
    focus_offset: float = 0.0


class Extraction:
    """Extract centroids and PSF information from an image."""

    __VALID_STAR_FINDER = ["daophot", "sextractor", "marginal"]

    def __init__(
        self,
        observatory: str,
        star_finder: str | None = None,
        pixel_scale: float | None = None,
        daophot_params={},
        marginal_params={},
        **params,
    ):

        self.observatory = observatory.upper()
        self.pixel_scale = pixel_scale or config["pixel_scale"]

        self.params = deepcopy(config["extraction"])
        self.params.update(params)
        self.params["daophot"].update(daophot_params)
        self.params["marginal"].update(marginal_params)

        self.output_dir = pathlib.Path(self.params["output_dir"])

        self.star_finder = star_finder or self.params["star_finder"]
        if self.star_finder not in self.__VALID_STAR_FINDER:
            raise ValueError(
                f"Invalid star finder. Valid values are {self.__VALID_STAR_FINDER}."
            )

    def process(self, image: PathLike) -> ExtractionData:
        """Process an image."""

        hdu = fits.open(image)
        data = hdu[1].data
        header = hdu[1].header

        camera = header["CAMNAME"][0:-1]  # Remove the n/s at the end of the camera name
        observatory = header["OBSERVAT"]

        obstime = Time(header["DATE-OBS"], format="iso", scale="tai")
        obstime += TimeDelta(header["EXPTIMEN"] / 2.0, format="sec")

        path = pathlib.Path(image)
        path = path.absolute()

        mjd = int(path.parts[-2])

        if match := re.match(r".*gimg\-gfa(\d)[ns]\-(\d+)\.fits", path.parts[-1]):
            cam_no = int(match.group(1))
            exp_no = int(match.group(2))
        else:
            cam_no = 0
            exp_no = 0

        if self.star_finder == "sextractor":
            regions = self._process_sextractor(data, path)[0]
        elif self.star_finder == "daophot":
            regions = self._process_daophot(data, path)
        elif self.star_finder == "marginal":
            regions = self._process_marginal(data, path)
        else:
            regions = pandas.DataFrame()

        regions["mjd"] = mjd
        regions["exposure"] = exp_no
        regions["camera"] = cam_no
        regions["fwhm_valid"] = 1

        if len(regions) > 0 and self.params["rejection_method"] is not None:
            self.reject(regions)

        valid = regions.loc[regions.fwhm_valid == 1]

        if len(valid) > 0:
            perc_50 = numpy.percentile(valid.fwhm, 50)
            fwhm_median = valid.loc[valid.fwhm < perc_50].fwhm.median()
            fwhm_median_round = float(numpy.round(fwhm_median, 3))

            # Prevent NaNs here since this is output to the headers.
            if numpy.isnan(fwhm_median_round):
                fwhm_median_round = -999.0
        else:
            fwhm_median_round = -999.0

        extraction_data = ExtractionData(
            str(image),
            path,
            camera,
            exp_no,
            int(obstime.mjd),
            obstime,
            observatory,
            field_ra=header["RAFIELD"],
            field_dec=header["DECFIELD"],
            field_pa=header["FIELDPA"],
            algorithm=self.star_finder,
            regions=regions,
            nregions=len(regions),
            nvalid=sum(regions.valid == 1),
            fwhm_median=fwhm_median_round,
            focus_offset=config["cameras"]["focus_offset"][camera],
        )

        output_file = self._get_output_path(path).with_suffix(".csv")
        output_file.unlink(missing_ok=True)

        regions.to_csv(str(output_file))

        return extraction_data

    def multiprocess(
        self,
        images: list[PathLike],
        n_cpus: int | None = None,
    ) -> list[ExtractionData]:

        n_cpus = n_cpus or multiprocessing.cpu_count()

        with multiprocessing.Pool(n_cpus) as pool:
            results = pool.map(self.process, images)

        return results

    def _get_output_path(self, path: pathlib.Path):
        """Returns the root path for output files."""

        path_no_suffix = path.with_suffix("")

        dirname = path_no_suffix.parent
        basename = path_no_suffix.parts[-1]

        if self.output_dir.is_absolute():
            output_dir = self.output_dir
        else:
            output_dir = dirname / self.output_dir

        output_dir.mkdir(parents=True, exist_ok=True)

        return output_dir / basename

    def _process_sextractor(
        self,
        data: numpy.ndarray,
        path: pathlib.Path,
        plot: bool | None = None,
    ):
        """Process image data using SExtractor."""

        if plot is None:
            plot = self.params.get("plot", False)

        output_root = self._get_output_path(path)

        back = sep.Background(data.astype("i4"))

        if plot:
            with plt.ioff():
                fig, ax = plt.subplots()
                ax.imshow(back, origin="lower")
                ax.set_title("Background: " + path.parts[-1])
                ax.grid(False)
                fig.savefig(str(output_root) + "-background.pdf")

        regions = sep.extract(
            data - back.back(),
            self.params["background_sigma"],
            err=back.globalrms,
        )

        regions = pandas.DataFrame(regions)
        regions.index.set_names("regions_id", inplace=True)
        regions.loc[:, "valid"] = 0

        min_npix = self.params["min_npix"]
        if len(regions) > 0:
            regions.loc[regions.npix > min_npix, "valid"] = 1

            # Keep only detectons with flux and sort them by highest flux first.
            regions.dropna(subset=["flux"], inplace=True)
            regions.sort_values("flux", inplace=True, ascending=False)

        if self.params["max_stars"]:
            regions = regions.loc[regions.valid == 1, :]
            regions = regions.head(self.params["max_stars"])

        regions = calculate_fwhm_from_ellipse(regions, self.pixel_scale)

        if plot:
            data_back = data - back.back()
            self.plot_regions(
                regions,
                data_back,
                path=path,
                vmin=data_back.mean() - back.globalrms,
                vmax=data_back.mean() + back.globalrms,
                factor=5.0,
                title=path.parts[-1] + " (SExtractor)",
            )

        return regions, back

    def _process_daophot(
        self,
        data: numpy.ndarray,
        path: pathlib.Path,
        plot: bool | None = None,
    ) -> pandas.DataFrame:
        """Process using iteratively subtracted PSF photometry."""

        if plot is None:
            plot = self.params.get("plot", False)

        psf_fwhm = self.params["daophot"]["initial_psf_fwhm"]
        psf_sigma = psf_fwhm / gaussian_sigma_to_fwhm

        background_sigma = self.params["background_sigma"]
        max_stars = self.params["daophot"].get("max_stars", None)

        data = data.astype("float32")

        bkgrms = StdBackgroundRMS()
        std = bkgrms(data)

        if self.params["daophot"].get("use_sep_finder", False):
            sep_regions = self._process_sextractor(data, path, plot=False)[0]
            initial_guesses = Table(
                names=["x_0", "y_0"],
                data=[sep_regions.x.iloc[0:max_stars], sep_regions.y.iloc[0:max_stars]],
            )
            finder = None
        else:
            initial_guesses = None
            finder = DAOStarFinder(
                threshold=background_sigma * std,
                fwhm=psf_fwhm,
                brightest=max_stars,
            )

        daogroup = DAOGroup(3.0 * psf_fwhm)

        mmm_bkg = MedianBackground()

        psf_model = IntegratedGaussianPRF(sigma=psf_sigma)
        psf_model.sigma.fixed = self.params["daophot"].get("fixed_sigma", False)

        fitter = LevMarLSQFitter()

        photometry = BasicPSFPhotometry(
            finder=finder,
            group_maker=daogroup,
            bkg_estimator=mmm_bkg,
            psf_model=psf_model,
            fitter=fitter,
            fitshape=(15, 15),
        )

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            regions = photometry(image=data, init_guesses=initial_guesses).to_pandas()

        regions["fwhm"] = regions.sigma_fit * gaussian_sigma_to_fwhm * self.pixel_scale
        regions["valid"] = 1

        if plot:
            self.plot_regions(
                regions,
                data,
                path=path,
                vmin=data.mean() - std,
                vmax=data.mean() + std,
                factor=5.0,
                xcen_col="x_0",
                ycen_col="y_0",
                a_col="sigma_fit",
                b_col="sigma_fit",
                theta_col=None,
                title=path.parts[-1] + " (DAOphot)",
            )

        return regions

    def _process_marginal(
        self,
        data: numpy.ndarray,
        path: pathlib.Path,
        plot: bool | None = None,
    ):
        """Extracts regions using the marginal distribution.

        Determines the initial centroids using the SExtractor routine.
        Then a box around each detection is selected and background-subtracted.
        The marginal distributions of the detection (the collapsed sum on each axis)
        are fitted using both a 1D Gaussian and 1D trapezoid, and the FWHM for each
        are calculated. Then the best match for each detection is selected based on
        the option that minimises the residuals.

        """

        marginal_params = self.params["marginal"]

        regions, back = self._process_sextractor(data, path, plot=False)
        regions = regions.loc[regions.valid == 1]
        regions = regions.rename(columns={"fwhm": "fwhm_sextractor"})

        regions["valid"] = 1 * (regions.npix > self.params["min_npix"])

        # Parameters of the Gaussian fit.
        regions["gaussian_fit"] = 1
        regions["x_gaussian"] = numpy.nan
        regions["y_gaussian"] = numpy.nan
        regions["fwhm_gaussian"] = numpy.nan
        regions["residual_gaussian"] = numpy.nan

        # Parameters of the trapezoid fit.
        regions["trapezoid_fit"] = 1
        regions["x_trapezoid"] = numpy.nan
        regions["y_trapezoid"] = numpy.nan
        regions["fwhm_trapezoid"] = numpy.nan
        regions["residual_trapezoid"] = numpy.nan

        # Parameters of the best fit. model_fit=g if the best fit was the Gaussian,
        # t if the trapezoid, and empty if invalid.
        regions["model_fit"] = ""
        regions["x_fit"] = numpy.nan
        regions["y_fit"] = numpy.nan
        regions["fwhm"] = numpy.nan
        regions["residual_fit"] = numpy.nan

        fitter = LevMarLSQFitter()
        fit_box = marginal_params["fit_box"]

        p_scale = self.pixel_scale
        data_back = data - back.back()

        for index, row in regions.iterrows():

            index = cast(Any, index)

            # Ignore detections that we have already marked as invalid in SExtractor.
            if row.valid == 0:
                continue

            cen = numpy.round([row.y, row.x]).astype(int)

            gauss_residual = []
            gauss_centroids = []
            gauss_fwhm = []
            gauss_valid = True

            trap_residual = []
            trap_centroids = []
            trap_fwhm = []
            trap_valid = True

            data_region = data_back[
                cen[0] - fit_box[0] // 2 : cen[0] + fit_box[0] // 2 + 1,
                cen[1] - fit_box[1] // 2 : cen[1] + fit_box[1] // 2 + 1,
            ].astype("f4")

            # Deal with regions near the edges
            if (
                data_region.shape[0] != data_region.shape[1]
                or data_region.shape[0] < fit_box[0]
                or data_region.shape[1] < fit_box[1]
            ):
                continue

            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                for axis in [0, 1]:
                    mid = fit_box[axis] // 2
                    x_mesh = numpy.arange(fit_box[axis])

                    data_marginal = data_region.sum(axis)
                    data_marginal /= data_marginal.max()

                    gauss_init = Gaussian1D(
                        amplitude=data_marginal.max(),
                        stddev=row.fwhm_sextractor / p_scale / gaussian_sigma_to_fwhm,
                        mean=mid,
                    )
                    g = fitter(gauss_init, x_mesh, data_marginal)
                    if gauss_valid is True:
                        # Reject if the fitting routine doesn't say this is a good fit.
                        gauss_valid = fitter.fit_info["ierr"] <= 4
                    if numpy.abs(g.mean - mid) > 10:
                        gauss_valid = False
                    gauss_centroids.append(cen[axis] + (g.mean - mid) + 0.5)
                    gauss_fwhm.append(g.stddev * gaussian_sigma_to_fwhm * p_scale)
                    gauss_residual.append(numpy.sum((data_marginal - g(x_mesh)) ** 2))

                    trap_init = Trapezoid1D(
                        amplitude=data_marginal.max(),
                        x_0=mid,
                        width=row.fwhm_sextractor / p_scale / gaussian_sigma_to_fwhm,
                        slope=0.1,
                    )
                    t = fitter(trap_init, x_mesh, data_marginal)
                    if trap_valid is True:
                        trap_valid = fitter.fit_info["ierr"] <= 4
                    if numpy.abs(t.x_0 - mid):
                        trap_valid = False
                    trap_centroids.append(cen[axis] + (t.x_0 - mid) + 0.5)
                    trap_fwhm.append((t.amplitude / t.slope + t.width) * p_scale)
                    trap_residual.append(numpy.sum((data_marginal - t(x_mesh)) ** 2))

            #         with plt.ioff():
            #             fig, ax = plt.subplots()
            #             ax.plot(data_marginal, c="k", ls="dotted")
            #             ax.plot(g(x_mesh), "b-")
            #             ax.plot(t(x_mesh), "g-")
            #             ax.axhline(y=0.0, ls="--", c="r")
            #             fig.savefig(
            #                 path.parent
            #                 / "extraction"
            #                 / (path.stem + f"-{index}-{axis}.pdf")
            #             )

            # plt.close("all")

            regions.loc[index, "gaussian_fit"] = int(gauss_valid)
            regions.loc[index, "x_gaussian"] = gauss_centroids[::-1][0]
            regions.loc[index, "y_gaussian"] = gauss_centroids[::-1][1]
            regions.loc[index, "fwhm_gaussian"] = float(numpy.mean(gauss_fwhm))
            gauss_residual_mean = (numpy.array(gauss_residual) ** 2).sum() ** 0.5
            regions.loc[index, "residual_gaussian"] = gauss_residual_mean

            regions.loc[index, "trapezoid_fit"] = int(trap_valid)
            regions.loc[index, "x_trapezoid"] = trap_centroids[::-1][0]
            regions.loc[index, "y_trapezoid"] = trap_centroids[::-1][1]
            regions.loc[index, "fwhm_trapezoid"] = float(numpy.mean(trap_fwhm))
            trap_residual_mean = (numpy.array(trap_residual) ** 2).sum() ** 0.5
            regions.loc[index, "residual_trapezoid"] = trap_residual_mean

            if not gauss_valid and not trap_valid:
                continue
            elif gauss_valid and gauss_residual_mean <= trap_residual_mean:
                regions.loc[index, "x_fit"] = gauss_centroids[::-1][0]
                regions.loc[index, "y_fit"] = gauss_centroids[::-1][1]
                regions.loc[index, "fwhm"] = float(numpy.mean(gauss_fwhm))
                regions.loc[index, "residual_fit"] = gauss_residual_mean
                regions.loc[index, "model_fit"] = "g"
            else:
                regions.loc[index, "x_fit"] = trap_centroids[::-1][0]
                regions.loc[index, "y_fit"] = trap_centroids[::-1][1]
                regions.loc[index, "fwhm"] = float(numpy.mean(trap_fwhm))
                regions.loc[index, "residual_fit"] = trap_residual_mean
                regions.loc[index, "model_fit"] = "t"

        regions.loc[regions.model_fit == "", "valid"] = 0

        if plot:
            self.plot_regions(
                regions.loc[regions.model_fit != ""],
                data,
                path=path,
                vmin=data.mean() - 5 * back.globalrms,
                vmax=data.mean() + 5 * back.globalrms,
                factor=1.0,
                xcen_col="x_fit",
                ycen_col="y_fit",
                a_col="fwhm",
                b_col="fwhm",
                theta_col=None,
                title=path.parts[-1] + " (marginal)",
            )

        return regions

    def reject(self, regions: pandas.DataFrame):
        """Rejects invalid FHWM measurements."""

        method = self.params.get("rejection_method", "sigclip")

        if self.star_finder in ["sextractor", "marginal"]:

            # Filter out bad FWHM values.
            ecc = numpy.sqrt(regions.a**2 - regions.b**2) / regions.a

            filter = (
                ((regions.a * self.pixel_scale) < 5)
                & ((regions.a * self.pixel_scale) > 0.4)
                & (regions.cpeak < 60000)
                & (ecc < 0.7)
            )

            regions.loc[~filter, "fwhm_valid"] = 0

        if method == "sigclip":
            fwhm = numpy.ma.array(
                regions.fwhm.values,
                mask=((regions.valid == 0) | (regions.fwhm_valid == 0)),
            )
            sigma = self.params.get("reject_sigma", 3.0)
            sigma_clip = SigmaClip(sigma, cenfunc="median")
            masked: Any = sigma_clip(fwhm, masked=True)
            regions.loc[masked.mask, "fwhm_valid"] = 0

        elif method == "nreject":
            nreject = self.params.get("nreject", 3)
            if isinstance(nreject, (int, float)):
                nreject = [int(nreject), int(nreject)]
            nreject = numpy.array(nreject, dtype=int)

            regions_sorted = regions.sort_values("fwhm", ascending=True)

            while True:
                # Endure we get at least some valid targets.
                valid = regions_sorted.iloc[nreject[0] : -nreject[1]]
                if len(valid) > 0:
                    break
                else:
                    nreject = nreject // 2

            regions.loc[~regions.index.isin(valid.index), "fwhm_valid"] = 0

        else:
            raise ValueError(f"Invalid rejection method {method!r}.")

        return regions

    def plot_regions(
        self,
        regions: pandas.DataFrame,
        data: numpy.ndarray,
        path: pathlib.Path,
        only_valid: bool = True,
        xcen_col: str = "x",
        ycen_col: str = "y",
        a_col: str = "a",
        b_col: str = "b",
        theta_col: str | None = "theta",
        factor: float = 1.0,
        vmin: float | None = None,
        vmax: float | None = None,
        title: str | None = None,
    ):
        """Plot regions."""

        with plt.ioff():
            fig, ax = plt.subplots()

        ax.set_title(title or path.parts[-1])
        ax.grid(False)

        # Image
        ax.imshow(data, origin="lower", cmap="gray", vmin=vmin, vmax=vmax)

        if only_valid:
            regions = regions.loc[regions.valid == 1]

        # Centroids
        ax.scatter(
            regions[xcen_col],
            regions[ycen_col],
            marker="x",  # type: ignore
            s=6,
            c="b",
            linewidths=0.5,
        )

        # Ellipses colour-coded by FWHM
        for _, region in regions.iterrows():
            ell = Ellipse(
                (region[xcen_col], region[ycen_col]),
                region[b_col] * 2 * factor,
                region[a_col] * 2 * factor,
                region[theta_col] if theta_col else 0.0,
                facecolor="None",
                edgecolor="r",
                linewidth=0.5,
            )
            ell.set_clip_box(ax.bbox)
            ax.add_patch(ell)

        ax.set_xlim(0, data.shape[1] - 1)
        ax.set_ylim(0, data.shape[0] - 1)

        output_root = self._get_output_path(path)
        fig.savefig(str(output_root) + "-centroids.pdf")


def calculate_fwhm_from_ellipse(
    regions: pandas.DataFrame,
    pixel_scale: float,
    a_col: str = "a",
    b_col: str = "b",
):
    """Calcualtes the FWHM from a list of SExtractor detections.

    Parameters
    ----------
    regions
        The pandas data frame with the list of regions. Usually an output from
        ``sep``. Must include columns ``valid``, ``a``, and ``b``.
    pixel_scale
        The pixel scale in arcsec per pixel.
    a_col
        Column with the semi-major axis measurement.
    b_col
        Column with the semi-minor axis measurement.

    Returns
    -------
    dataframe
        The input ``regions`` data frame with an additional column ``fwhm``
        calculated from the ellipse parameters

    """

    fwhm_pixel = 2 * (numpy.log(2) * (regions[a_col] ** 2 + regions[b_col] ** 2)) ** 0.5
    fwhm = pixel_scale * fwhm_pixel

    regions = regions.copy()
    regions["fwhm"] = fwhm

    return regions



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2020-01-29
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import signal
from contextlib import suppress

import click
from click_default_group import DefaultGroup

from clu.tools import cli_coro
from sdsstools.daemonizer import DaemonGroup

from cherno.actor.actor import ChernoActor


async def shutdown(signal, loop, actor):
    """Cancel tasks, including run_forever()."""

    tasks = [t for t in asyncio.all_tasks() if t is not asyncio.current_task()]
    [task.cancel() for task in tasks]

    with suppress(asyncio.CancelledError):
        await asyncio.gather(*tasks)


@click.group(cls=DefaultGroup, default="actor", default_if_no_args=True)
@click.option(
    "-o",
    "--observatory",
    type=str,
    help="Observatory configuration to use. Defaults to using $OBSERVATORY.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode. Use additional v for more details.",
)
@click.pass_context
def cherno(ctx, verbose: bool = False, observatory: str | None = None):
    """Cherno CLI."""

    ctx.obj = {"verbose": verbose, "observatory": observatory}


@cherno.group(cls=DaemonGroup, prog="cherno_actor", workdir=os.getcwd())
@click.pass_context
@cli_coro
async def actor(ctx):
    """Runs the actor."""

    observatory = ctx.obj["observatory"] or os.environ["OBSERVATORY"]

    config_file = os.path.join(
        os.path.dirname(__file__),
        f"etc/cherno_{observatory}.yml",
    )

    cherno_actor = ChernoActor.from_config(config_file)

    loop = asyncio.get_event_loop()
    signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT)
    for s in signals:
        loop.add_signal_handler(
            s,
            lambda s=s: asyncio.create_task(shutdown(s, loop, cherno_actor)),
        )

    try:
        await cherno_actor.start()
        await cherno_actor.run_forever()
    except asyncio.CancelledError:
        pass
    finally:
        await cherno_actor.stop()
        loop.stop()


if __name__ == "__main__":
    cherno()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-16
# @Filename: show.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from pprint import pformat

from . import ChernoCommandType, cherno_parser


__all__ = ["show"]


@cherno_parser.command()
async def show(command: ChernoCommandType):
    """Shows current configuration options."""

    assert command.actor

    skip_fields = ["actor", "camera_state", "scale_history"]

    for field in command.actor.state.__dataclass_fields__:
        if field in skip_fields or field.startswith("_"):
            continue
        lines = pformat(getattr(command.actor.state, field), width=60).splitlines()
        for nline, line in enumerate(lines):
            if nline == 0:
                command.info(text=f"{field}: {line}")
            else:
                spaces = " " * len(field)
                command.info(text=f"{spaces}  {line}")

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-30
# @Filename: guide.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import click

from cherno.maskbits import GuiderStatus

from . import ChernoCommandType, cherno_parser


__all__ = ["guide"]


@cherno_parser.command()
@click.option(
    "-o",
    "--off",
    is_flag=True,
    help="Turns guiding off.",
)
@click.option(
    "-t",
    "--timeout",
    type=float,
    default=5.0,
    help="Extra time to wait for an exposure to be done before failiing.",
)
async def guide(command: ChernoCommandType, off: bool = False, timeout: float = 5.0):
    """Starts the guiding loop."""

    assert command.actor

    state = command.actor.state

    if off:
        if state.status & GuiderStatus.STOPPING:
            return command.fail(error="Guider is already stopping.")
        elif not (state.status & GuiderStatus.EXPOSING):
            return command.fail(error="Guider is not exposing.")

    command.info("Starting guider loop.")

    while True:
        if state.status & (GuiderStatus.STOPPING | GuiderStatus.FAILED):
            return command.finish("Finishing guider loop.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-09
# @Filename: exposer.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import os
import re
from glob import glob

from typing import Any, Callable, Union

from astropy.time import Time

from clu import Command

from cherno import config
from cherno.exceptions import ExposerError
from cherno.maskbits import GuiderStatus

from . import ChernoCommandType


CallbackType = Union[Callable[[ChernoCommandType, list[str]], Any], None]


class Exposer:
    """Helper to expose the cameras.

    Parameters
    ----------
    command
        Actor command requesting the exposure loop.
    callback
        A callback to invoke every time a group of exposures finishes. The
        callback is invoked with the command and a list of new exposure files.
        If the callback is a coroutine it is run in a task if blocking is
        disabled and awaited otherwise. Use `.set_blocking` to change the
        blocking behaviour. The default behaviour is to block. If the callback
        is a function, it is run in an executor.

    """

    def __init__(self, command: ChernoCommandType, callback: CallbackType = None):

        self.command = command

        assert command.actor
        self.actor = command.actor

        self.actor_state = self.actor.state

        self.callback = callback
        self._blocking: bool = True

    def set_blocking(self, blocking: bool):
        """Whether to block while calling the callback."""

        self._blocking = blocking

    def fail(self, message=""):
        """Sets the guider status to failed and raises an exception."""

        self.actor_state.set_status(GuiderStatus.FAILED | GuiderStatus.IDLE)
        raise ExposerError(message)

    async def one(self, exposure_time: float, **kwargs):
        """Exposes the cameras once.

        Parameters
        ----------
        exposure_time
            The exposure time to command.
        kwargs
            Arguments to pass to `.loop`.

        """

        if "count" in kwargs:
            raise ExposerError("Cannot specify count with one().")

        return await self.loop(exposure_time, count=1, **kwargs)

    def is_stopping(self):
        """Is the guider loop stopping?"""

        return (self.actor_state.status & GuiderStatus.STOPPING).value > 0

    async def loop(
        self,
        exposure_time: float | None = None,
        count: int | None = None,
        delay: float = 0.0,
        names: list[str] | None = None,
        timeout: float | None = None,
        callback: CallbackType = None,
    ):
        """Loops the cameras.

        Parameters
        ----------
        exposure_time
            The exposure time to command. If `None`, uses the stored exposure
            time in the actor state.
        count
            The number of exposures to take before stopping the loop.
        delay
            How long to wait between the end of an exposure and the beginning
            of the next one.
        names
            FLISwarm camera names to expose. If `None`, uses the cameras specified
            in the configuration.
        timeout
            How long to wait until the end of the exposure time for the FLISwarm
            command to finish before cancelling the loop.
        callback
            The callback to invoke after each exposure completes. If specified,
            overrides the global callback only for this loop.

        """

        if self.actor.tron is None:
            raise ExposerError("Tron is not connected. Cannot expose.")

        if (self.actor_state.status & GuiderStatus.NON_IDLE).value > 0:
            raise ExposerError("The guider is already exposing.")

        if exposure_time is not None:
            self.actor_state.exposure_time = exposure_time

        n_exp = 0
        while True:

            if self.is_stopping() or (count is not None and n_exp >= count):
                self.actor_state.set_status(GuiderStatus.STOPPING, mode="remove")
                self.actor_state.set_status(GuiderStatus.IDLE, mode="add")
                return

            self._check_ffs()

            # Clear the FAILED flag if it's set
            self.actor_state.set_status(GuiderStatus.FAILED, mode="remove")

            # Set the status of the guider as EXPOSING. This removes IDLE.
            self.actor_state.set_status(GuiderStatus.EXPOSING, mode="add")

            names = names or self.actor.state.enabled_cameras
            if names is None or len(names) == 0:
                self.fail("No cameras defined.")

            num = self._get_num(names)

            names_comma = ",".join(names)

            # Issue a fliswarm talk status to the cameras to update their status.
            try:
                status_command = await asyncio.wait_for(
                    self.actor.tron.send_command(
                        "fliswarm",
                        f"talk -n {names_comma} status",
                    ),
                    3,
                )
            except asyncio.TimeoutError:
                self.fail("Timed out updating camera status.")

            if status_command.status.did_fail:
                self.fail("Failed updating camera status.")

            try:
                etime = self.actor_state.exposure_time
                expose_command = await asyncio.wait_for(
                    self.actor.tron.send_command(
                        "fliswarm",
                        f"talk -n {names_comma} -- expose -n {num} {etime}",
                    ),
                    etime + timeout if timeout is not None else None,
                )
            except asyncio.TimeoutError:
                self.fail("Timed out waiting for the exposure to finish.")

            if expose_command.status.did_fail:
                self.fail("Expose command failed.")

            self.actor_state.set_status(GuiderStatus.EXPOSING, mode="remove")

            if self.is_stopping():
                # Continue, the first check in the new iteration will stop the loop.
                continue

            filenames = self._get_filename_bundle(expose_command)
            if filenames is None:
                self.fail("The keyword filename_bundle was not output.")
            elif len(filenames) == 0:
                self.fail("The keyword filename_bundle is empty.")
            else:
                try:
                    await self.invoke_callback(
                        filenames,
                        callback=callback or self.callback,
                    )
                except Exception as err:
                    self.fail(str(err))

            if delay:
                await asyncio.sleep(delay)

            n_exp += 1

            self.actor_state.set_status(GuiderStatus.IDLE, mode="add")

    def _check_ffs(self):
        """Checks that the FFS are open."""

        if self.actor.observatory == "LCO":
            return

        values = self.actor.models["mcp"]["ffsStatus"].value
        if len(values) == 0 or all([value is None for value in values]):
            self.command.warning("FFS status unknown.")

        if not all([int(ss) == 10 for ss in values]):
            self.command.warning("FFS petals are not open.")

    def _get_num(self, names: list[str]) -> int:
        """Returns the next sequence number."""

        date = Time.now()
        mjd = int(date.mjd)

        dirpath = os.path.join(config["cameras"]["path"], str(mjd))
        if not os.path.exists(dirpath):
            return 1

        gimgs = glob(os.path.join(dirpath, "*.fits*"))
        matches = [
            int(m.group(1))
            for file_ in gimgs
            if (m := re.search(r"gimg\-.+?\-([0-9]+)", file_))
        ]

        return max(set(matches)) + 1

    def _get_filename_bundle(self, command: Command):
        """Returns the ``filename_bundle`` values from the list of command replies."""

        for reply in command.replies:
            for reply_key in reply.keywords:
                key_name = reply_key.name.lower()
                if key_name != "filename_bundle":
                    continue
                filenames = [value.native for value in reply_key.values]
                return filenames

        return None

    async def invoke_callback(
        self,
        filenames: list[str],
        callback: CallbackType = None,
    ):
        """Invokes the callback with a list of filenames."""

        def unset_processing(*args, **kwargs):
            """Removes the ``PROCESSING`` flag after the callback completes."""
            self.actor_state.set_status(GuiderStatus.PROCESSING, mode="remove")

        self.actor_state.set_status(GuiderStatus.PROCESSING, mode="add")

        callback = callback or self.callback
        if callback is None:
            self.command.warning("Exposer: no callback defined.")
            return

        if asyncio.iscoroutinefunction(callback) or (
            hasattr(callback, "func") and asyncio.iscoroutinefunction(callback.func)
        ):
            task = asyncio.create_task(callback(self.command, filenames))
        else:
            task = asyncio.get_running_loop().run_in_executor(
                None,
                callback,
                self.command,
                filenames,
            )

        task.add_done_callback(unset_processing)

        if self._blocking:
            await task

        return



# !usr/bin/env python
# -*- coding: utf-8 -*-
#
# Licensed under a 3-clause BSD license.
#
# @Author: Brian Cherinka
# @Date:   2017-12-05 12:01:21
# @Last modified by:   Brian Cherinka
# @Last Modified time: 2017-12-05 12:19:32


class ChernoError(Exception):
    """A custom core Cherno exception"""

    def __init__(self, message=None):

        message = "There has been an error" if not message else message

        super(ChernoError, self).__init__(message)


class ExposerError(ChernoError):
    """An error in the `.Exposer` class."""

    pass


class ChernoNotImplemented(ChernoError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):

        message = "This feature is not implemented yet." if not message else message

        super(ChernoNotImplemented, self).__init__(message)


class ChernoMissingDependency(ChernoError):
    """A custom exception for missing dependencies."""

    pass


class ChernoWarning(Warning):
    """Base warning for Cherno."""


class ChernoUserWarning(UserWarning, ChernoWarning):
    """The primary warning class."""

    pass


class ChernoDeprecationWarning(ChernoUserWarning):
    """A warning for deprecated features."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-09-01
# @Filename: converge.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from datetime import datetime

import click
import numpy

from clu.parsers.click import (
    cancel_command,
    get_current_command_name,
    get_running_tasks,
    timeout,
)

from . import ChernoCommandType, cherno_parser


__all__ = ["converge"]


@cherno_parser.command()
@click.argument("RMS", type=float)
@click.option(
    "-n",
    "--iterations",
    type=int,
    default=2,
    help="Number of guide loop iterations the RMS must be below the required level.",
)
@click.option(
    "--max-age",
    type=float,
    default=300,
    help="Maximum age, in seconds, of the RMS mesurements.",
)
@click.option(
    "--require-new",
    is_flag=True,
    help="Requires that the last RMS measurement is less than 5 seconds old.",
)
@timeout(600)
async def converge(
    command: ChernoCommandType,
    rms: float | None = None,
    iterations: int = 2,
    max_age: float = 300,
    require_new: bool = True,
):
    """Runs until the guide RMS has been below a given threshold for N iterations."""

    tasks = get_running_tasks(get_current_command_name())
    if tasks and len(tasks) > 1:
        command.warning("Another version of this command was running. Cancelling it.")
        cancel_command(keep_last=True)

    assert command.actor.store is not None

    while True:
        kws: list = command.actor.store.tail("guide_rms", n=iterations)

        kws = [k for k in kws if (datetime.now() - k.date).total_seconds() < max_age]
        values = numpy.array([k.value[-1] for k in kws if len(k.value) > 0])

        last_age = (datetime.now() - kws[-1].date).total_seconds()
        if (not require_new) or (require_new and last_age < 5):
            if len(values) >= iterations and numpy.all(values <= rms):
                return command.finish("RMS has been reached.")

        await asyncio.sleep(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-19
# @Filename: status.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import pathlib

from yaml import warnings

from cherno import config
from cherno.exceptions import ChernoUserWarning

from . import ChernoCommandType, cherno_parser


__all__ = ["status"]


def get_astrometrynet_paths():
    """Returns a list of astrometry.net index paths that cherno will use."""

    astrometry_net_config = config["acquisition"]["astrometry_net_config"]
    backend_config = pathlib.Path(__file__).parents[1] / astrometry_net_config

    if not backend_config.exists():
        warnings.warn(
            "Cannot find astrometry.net backend configuration. "
            "This is a critical issue.",
            ChernoUserWarning,
        )
        return []

    index_paths: list[str] = []
    for line in open(backend_config, "r").readlines():
        line = line.strip()
        if line.startswith("#"):
            continue
        if line.startswith("add_path"):
            index_paths.append(line.split()[1])

    return index_paths


@cherno_parser.command()
async def status(command: ChernoCommandType):
    """Reports the status of the system."""

    command.info(guider_status=hex(command.actor.state.status.value))
    command.info(enabled_axes=command.actor.state.enabled_axes)

    command.info(astrometrynet_index_paths=get_astrometrynet_paths())

    default_offset = config.get("default_offset", (0.0, 0.0, 0.0))
    command.info(default_offset=default_offset)
    command.info(offset=command.actor.state.offset)

    for axis in ["ra", "dec", "rot", "focus"]:
        command.info(
            message={
                f"pid_{axis}": list(
                    command.actor.state.guide_loop[axis]["pid"].values()
                )
            }
        )

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-12
# @Filename: stop.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from cherno.maskbits import GuiderStatus

from . import ChernoCommandType, cherno_parser


__all__ = ["stop"]


@cherno_parser.command()
async def stop(command: ChernoCommandType):
    """Stops any running guide loops."""

    assert command.actor is not None

    if command.actor.state.status & GuiderStatus.IDLE:
        command.warning("The guider is idle.")
        return command.finish()

    command.actor.state.set_status(GuiderStatus.STOPPING, mode="add")

    return command.finish("The guide loop is stopping.")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cherno', 'cherno.actor']

package_data = \
{'': ['*'], 'cherno': ['etc/*']}

install_requires = \
['astropy>=4.0.0',
 'click-default-group>=1.2.2,<2.0.0',
 'matplotlib>=3.1.1',
 'numpy>=1.19.2,<2.0.0',
 'pandas>=1.3.4,<2.0.0',
 'photutils>=1.3.0,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'sdss-clu>=1.8.0,<2.0.0',
 'sdss-coordio>=1.4.1',
 'sdssdb>=0.4.3',
 'sdsstools>=0.1.0',
 'seaborn>=0.11.2,<0.12.0',
 'sep>=1.2.0,<2.0.0',
 'simple-pid>=1.0.1,<2.0.0',
 'tables>=3.6.1']

entry_points = \
{'console_scripts': ['cherno = cherno.__main__:cherno']}

setup_kwargs = {
    'name': 'sdss-cherno',
    'version': '0.3.2',
    'description': 'SDSS guider actor',
    'long_description': '# cherno\n\n![Versions](https://img.shields.io/badge/python->3.8-blue)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Documentation Status](https://readthedocs.org/projects/sdss-cherno/badge/?version=latest)](https://sdss-cherno.readthedocs.io/en/latest/?badge=latest)\n[![Build](https://img.shields.io/github/workflow/status/sdss/cherno/Test)](https://github.com/sdss/cherno/actions)\n[![codecov](https://codecov.io/gh/sdss/cherno/branch/main/graph/badge.svg)](https://codecov.io/gh/sdss/cherno)\n\n\nSDSS guider actor\n\n## Installation\n\nIn general you should be able to install ``cherno`` by doing\n\n```console\npip install sdss-cherno\n```\n\nTo build from source, use\n\n```console\ngit clone git@github.com:sdss/cherno\ncd cherno\npip install .\n```\n\n## Development\n\n`cherno` uses [poetry](http://poetry.eustace.io/) for dependency management and packaging. To work with an editable install it\'s recommended that you setup `poetry` and install `cherno` in a virtual environment by doing\n\n```console\npoetry install\n```\n\n### Style and type checking\n\nThis project uses the [black](https://github.com/psf/black) code style with 88-character line lengths for code and docstrings. It is recommended that you run `black` on save. Imports must be sorted using [isort](https://pycqa.github.io/isort/). The GitHub test workflow checks all the Python file to make sure they comply with the black formatting.\n\nConfiguration files for [flake8](https://flake8.pycqa.org/en/latest/), [isort](https://pycqa.github.io/isort/), and [black](https://github.com/psf/black) are provided and will be applied by most editors. For Visual Studio Code, the following project file is compatible with the project configuration:\n\n```json\n{\n    "python.formatting.provider": "black",\n    "[python]" : {\n        "editor.codeActionsOnSave": {\n            "source.organizeImports": true\n        },\n        "editor.formatOnSave": true\n    },\n    "[markdown]": {\n        "editor.wordWrapColumn": 88\n    },\n    "[restructuredtext]": {\n        "editor.wordWrapColumn": 88\n    },\n    "editor.rulers": [88],\n    "editor.wordWrapColumn": 88,\n    "python.analysis.typeCheckingMode": "basic"\n}\n```\n\nThis assumes that the [Python](https://marketplace.visualstudio.com/items?itemName=ms-python.python) and [Pylance](https://marketplace.visualstudio.com/items?itemName=ms-python.vscode-pylance) extensions are installed.\n\nThis project uses [type hints](https://docs.python.org/3/library/typing.html). Typing is enforced by the test workflow using [pyright](https://github.com/microsoft/pyright) (in practice this means that if ``Pylance`` doesn\'t produce any errors in basic mode, ``pyright`` shouldn\'t).\n',
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sdss/cherno',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-11-30
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from typing import Union

from clu import Command, FakeCommand
from clu.parsers.click import command_parser as cherno_parser

from .actor import ChernoActor


ChernoCommandType = Union[Command[ChernoActor], FakeCommand]


from .acquire import *
from .converge import *
from .guide import *
from .offset import *
from .scale import *
from .set import *
from .show import *
from .status import *
from .stop import *
from .version import *



# encoding: utf-8

import os
import pathlib
import warnings

from typing import cast

from sdsstools import get_logger, get_package_version
from sdsstools.configuration import __ENVVARS__, get_config


NAME = "sdss-cherno"

# Inits the logging system. Only shell logging, and exception and warning catching.
# File logging can be started by calling log.start_file_logger(path).
log = get_logger(NAME)


__ENVVARS__["OBSERVATORY"] = "?"

OBSERVATORY = os.environ.get("OBSERVATORY", None)
if OBSERVATORY is None:
    OBSERVATORY = "APO"
    warnings.warn("Unknown observatory. Defaulting to APO!", UserWarning)
    config = get_config("cherno")
else:
    config = get_config(f"cherno_{OBSERVATORY}")


__version__ = cast(str, get_package_version(path=__file__, package_name=NAME))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-06
# @Filename: acquire.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from functools import partial

from typing import TYPE_CHECKING

import click

from cherno import config
from cherno.acquisition import Acquisition
from cherno.actor.exposer import Exposer
from cherno.exceptions import ExposerError

from . import cherno_parser


if TYPE_CHECKING:
    from . import ChernoCommandType


__all__ = ["acquire"]


@cherno_parser.command()
@click.option(
    "-t",
    "--exposure-time",
    type=float,
    default=None,
    help="Cameras exposure time.",
)
@click.option(
    "-m",
    "--cameras",
    type=str,
    help="Comma-separated cameras to expose.",
)
@click.option(
    "-c",
    "--continuous",
    is_flag=True,
    help="Run acquisition in continuous mode.",
)
@click.option(
    "-n",
    "--count",
    type=int,
    help="Number of acquisition iterations. Incompatible with --continuous.",
)
@click.option(
    "--apply/--no-apply",
    default=True,
    help="Whether to apply the correction.",
)
@click.option(
    "-f",
    "--full",
    is_flag=True,
    help="Applies the full correction once. Cannot be used with --continuous.",
)
@click.option(
    "--only-radec",
    is_flag=True,
    help="Only fits and corrects RA/Dec. Rotation and scale values are informative.",
)
@click.option(
    "--auto-radec-min",
    type=int,
    default=config["acquisition"]["auto_radec_min"],
    help="Number of cameras solving below which only RA/Dec will be fit. "
    "-1 disables this feature and a full fit is allways performed.",
)
@click.option(
    "-w",
    "--wait",
    type=float,
    help="Time to wait between iterations.",
)
@click.option(
    "--no-block",
    is_flag=True,
    help="Returns immediatly after starting the acquisition loop.",
)
async def acquire(
    command: ChernoCommandType,
    exposure_time: float | None = None,
    continuous: bool = False,
    count: int | None = None,
    apply: bool = True,
    full: bool = False,
    only_radec: bool = False,
    auto_radec_min: int = config["acquisition"]["auto_radec_min"],
    cameras: str | None = None,
    wait: float | None = None,
    no_block: bool = False,
):
    """Runs the acquisition procedure."""

    assert command.actor

    if count is not None and continuous is True:
        return command.fail("--count and --continuous are incompatible.")

    _exposure_loop = command.actor.state._exposure_loop
    if _exposure_loop is not None and not _exposure_loop.done():
        command.warning("An active Exposer loop was found. Cancelling it.")
        _exposure_loop.cancel()

    count = count or 1

    if cameras is None:
        names = None
    else:
        names = list(cameras.split(","))

    if exposure_time is not None:
        if exposure_time < 1.0:
            return command.fail("Exposure time not set or too small.")
        else:
            command.actor.state.exposure_time = exposure_time

    acquisition = Acquisition(config["observatory"])
    command.actor.state._acquisition_obj = acquisition  # To update PID coeffs.

    callback = partial(
        acquisition.process,
        correct=apply,
        full_correction=full,
        wait_for_correction=(wait is None),
        only_radec=only_radec,
        auto_radec_min=auto_radec_min,
    )
    exposer = Exposer(command, callback=callback)

    command.actor.state._exposure_loop = asyncio.create_task(
        exposer.loop(
            None,
            count=count if continuous is False else None,
            timeout=25,
            names=names,
            delay=wait or 0.0,
        )
    )

    if not no_block:
        try:
            await command.actor.state._exposure_loop
        except ExposerError as err:
            return command.fail(f"Acquisition failed: {err}")
    else:
        return command.finish("Finishing due to --no-block. The guide loop is running.")

    return command.finish()
